package org.processmining.plugins.petrinet.behavioralanalysis.woflan;

import java.util.Collection;
import java.util.HashSet;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;

import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.connections.ConnectionManager;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.collection.MultiSet;
import org.processmining.models.connections.petrinets.behavioral.DeadTransitionsConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.NonLiveSequencesConnection;
import org.processmining.models.connections.petrinets.behavioral.NonLiveTransitionsConnection;
import org.processmining.models.connections.petrinets.behavioral.UnboundedPlacesConnection;
import org.processmining.models.connections.petrinets.behavioral.UnboundedSequencesConnection;
import org.processmining.models.connections.petrinets.structural.AbstractComponentSetConnection;
import org.processmining.models.connections.petrinets.structural.AbstractInvariantMarkingConnection;
import org.processmining.models.connections.petrinets.structural.AbstractNetHandleConnection;
import org.processmining.models.connections.petrinets.structural.NonExtendedFreeChoiceClustersConnection;
import org.processmining.models.connections.petrinets.structural.NotPCoveredNodesConnection;
import org.processmining.models.connections.petrinets.structural.NotSCoveredNodesConnection;
import org.processmining.models.connections.petrinets.structural.PTHandleConnection;
import org.processmining.models.connections.petrinets.structural.PlaceInvariantConnection;
import org.processmining.models.connections.petrinets.structural.SComponentConnection;
import org.processmining.models.connections.petrinets.structural.ShortCircuitedNetConnection;
import org.processmining.models.connections.petrinets.structural.TPHandleConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.analysis.DeadTransitionsSet;
import org.processmining.models.graphbased.directed.petrinet.analysis.NonExtendedFreeChoiceClustersSet;
import org.processmining.models.graphbased.directed.petrinet.analysis.NonLiveSequences;
import org.processmining.models.graphbased.directed.petrinet.analysis.NonLiveTransitionsSet;
import org.processmining.models.graphbased.directed.petrinet.analysis.NotPCoveredNodesSet;
import org.processmining.models.graphbased.directed.petrinet.analysis.NotSCoveredNodesSet;
import org.processmining.models.graphbased.directed.petrinet.analysis.PTHandles;
import org.processmining.models.graphbased.directed.petrinet.analysis.PlaceInvariantSet;
import org.processmining.models.graphbased.directed.petrinet.analysis.SComponentSet;
import org.processmining.models.graphbased.directed.petrinet.analysis.TPHandles;
import org.processmining.models.graphbased.directed.petrinet.analysis.UnboundedPlacesSet;
import org.processmining.models.graphbased.directed.petrinet.analysis.UnboundedSequences;
import org.processmining.models.graphbased.directed.petrinet.analysis.WorkflowNetUtils;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.semantics.petrinet.Marking;

/**
 * Woflan
 * 
 * Woflan diagnosis for a Petri net.
 * 
 * @author HVERBEEK
 * 
 */
@Plugin(name = "Woflan", returnLabels = { "Woflan Diagnosis" }, returnTypes = { WoflanDiagnosis.class }, parameterLabels = { "Net" /*
																																	 * ,
																																	 * "Assumptions"
																																	 */}, userAccessible = true)
public class Woflan {

	/**
	 * The current plug-in context.
	 */
	private PluginContext context;

	/**
	 * The connection manager of this context.
	 */
	private ConnectionManager cm;

	/**
	 * The current diagnostic results.
	 */
	private WoflanDiagnosis diagnosis;

	/**
	 * The current state of the diagnosis: to start.
	 */
	private WoflanState state = WoflanState.INIT;

	private Collection<WoflanState> assumptions = new HashSet<WoflanState>();

	/**
	 * The initial marking of the short-circuited net. Note that the net and the
	 * short-circuited net are kept by the diagnosis.
	 * 
	 * Invariant: marking == null || marking == diagnosis.shortCNet
	 */
	private Marking initialMarking;
	private Marking finalMarking;

	/**
	 * Public constructor of Woflan
	 */
	public Woflan() {
		initialMarking = null;
		finalMarking = null;
	}

	/**
	 * The Woflan plug-in using no assumptions.
	 * 
	 * @param context
	 *            The context of this plug-in.
	 * @param net
	 *            The net to diagnose.
	 * @return The diagnosis of the net.
	 * @throws ExecutionException
	 * @throws InterruptedException
	 * @throws CancellationException
	 * @throws Exception
	 */
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "H.M.W. Verbeek", email = "h.m.w.verbeek@tue.nl")
	@PluginVariant(variantLabel = "Woflan", requiredParameterLabels = { 0 })
	public WoflanDiagnosis diagnose(PluginContext context, Petrinet net) throws Exception {
		return diagnose(context, net, new HashSet<WoflanState>());
	}

	/**
	 * The Woflan plug-in using assumptions.
	 * 
	 * @param context
	 *            The context of this plug-in.
	 * @param net
	 *            The net to diagnose.
	 * @param assumptions
	 *            The assumptions to make.
	 * @return The diagnosis of the net, given the assumptions.
	 * @throws CancellationException
	 * @throws InterruptedException
	 * @throws ExecutionException
	 */
	//TODO: BVD: A plugin variant cannot accept parameters using generics. At runtime,
	// these generics are lost, i.e. every Collection would be OK as a parameter, such as, 
	// for example a Marking, which is a MultiSet<Place> which extends Collection<Place> 
	// A Special wrapping object has to be defined for Collection<WoflanState>	
	//	
	//	@PluginVariant(variantLabel = "Woflan", requiredParameterLabels = { 0, 1 })
	public WoflanDiagnosis diagnose(PluginContext context, Petrinet net, Collection<WoflanState> assumptions)
			throws ConnectionCannotBeObtained {
		/**
		 * Store the context and connection manager for future use.
		 */
		this.context = context;
		cm = context.getConnectionManager();
		/**
		 * Store the assumptions for future use.
		 */
		this.assumptions = assumptions;
		/**
		 * Initialize the diagnosis.
		 */
		diagnosis = null;
		/**
		 * Try whether a diagnosis already exists for this net and these
		 * assumptions.
		 */
		try {
			for (WoflanConnection conn : cm.getConnections(WoflanConnection.class, context, net)) {
				WoflanDiagnosis object = conn.getObjectWithRole(WoflanConnection.DIAGNOSIS);
				if (conn.getAssumptions().equals(assumptions)) {
					/**
					 * Yes, it exists. Use it.
					 */
					diagnosis = object;
				}
			}
		} catch (Exception e) {
			/**
			 * No harm done.
			 */
		}
		if (diagnosis == null) {
			/**
			 * No, it does not exist. create a new one.
			 */
			diagnosis = new WoflanDiagnosis(net);
			/**
			 * Do the diagnosis.
			 */
			while (state != WoflanState.DONE) {
				state = diagnose(state);
			}
			if (context.getResult() != null && context.getResult().getSize() > 0) {
				context.getFutureResult(0).setLabel("Woflan Diagnosis of net " + net.getLabel());
			}
			context.log(diagnosis.toString());
			context.addConnection(new WoflanConnection(net, diagnosis, assumptions));
		}
		return diagnosis;
	}

	/**
	 * Takes a next diagnosis step.
	 * 
	 * @param state
	 *            The current state.
	 * @return The next state.
	 * @throws ExecutionException
	 * @throws InterruptedException
	 * @throws CancellationException
	 * @throws Exception
	 */
	public WoflanState diagnose(WoflanState state) throws ConnectionCannotBeObtained {
		switch (state) {
			case BOUNDED : {
				state = diagnoseDeadness();
				break;
			}
			case FREECHOICE : {
				state = diagnosePCover();
				break;
			}
			case INIT : {
				state = diagnoseWFNet();
				break;
			}
			case LIVE : {
				state = WoflanState.SOUND;
				break;
			}
			case NONFREECHOICE : {
				state = diagnosePTHandles();
				break;
			}
			case NOPCOVER : {
				state = diagnoseBoundedness();
				break;
			}
			case NOSCOVER : {
				state = diagnoseFC();
				break;
			}
			case NOTDEAD : {
				state = diagnoseLiveness();
				break;
			}
			case NOTWELLPTHANDLED : {
				state = diagnoseTPHandles();
				break;
			}
			case NOTWELLTPHANDLED : {
				state = diagnosePCover();
				break;
			}
			case PCOVER : {
				state = WoflanState.BOUNDED;
				break;
			}
			case SCOVER : {
				state = WoflanState.BOUNDED;
				break;
			}
			case SOUND : {
				state = WoflanState.DONE;
				diagnosis.setSound();
				break;
			}
			case WELLPTHANDLED : {
				state = diagnoseTPHandles();
				break;
			}
			case WELLTPHANDLED : {
				state = diagnosePCover();
				break;
			}
			case WFNET : {
				state = diagnoseSCover();
				break;
			}
			default : {
				state = WoflanState.DONE;
				break;
			}
		}
		return state;
	}

	/**
	 * Returns a copy of the net which has an extra transition with (1) all sink
	 * places as inputs and (2) all source places as output. If such a copy
	 * cannot be found, it is created and returned. Otherwise, an existing copy
	 * is returned.
	 */
	private void shortCircuit() {
		diagnosis.shortCNet = null;
		initialMarking = null;
		finalMarking = null;
		/**
		 * Try to find an existing copy.
		 */
		try {
			for (ShortCircuitedNetConnection conn : cm.getConnections(ShortCircuitedNetConnection.class, context,
					diagnosis.net)) {
				diagnosis.shortCNet = conn.getObjectWithRole(ShortCircuitedNetConnection.SCNET);
				finalMarking = conn.getObjectWithRole(ShortCircuitedNetConnection.FINALMARKING);

			}
			if (diagnosis.shortCNet != null) {
				for (InitialMarkingConnection conn : cm.getConnections(InitialMarkingConnection.class, context,
						diagnosis.shortCNet)) {
					initialMarking = conn.getObjectWithRole(InitialMarkingConnection.MARKING);

				}
			}
		} catch (Exception e) {
			/**
			 * No harm done, just continue.
			 */
		}
		/**
		 * If marking != null, the diagnosis.shortCNet != null.
		 */
		if (initialMarking == null) {
			/**
			 * No existing copy found. Create a copy.
			 */
			Object[] objects = WorkflowNetUtils.shortCircuit(context, diagnosis.net);
			diagnosis.shortCNet = (Petrinet) objects[0];
			initialMarking = (Marking) objects[1];
			finalMarking = (Marking) objects[2];
			/**
			 * Add the copy to the object pool.
			 */
			context.getProvidedObjectManager().createProvidedObject(
					"Short-circuited net of " + diagnosis.net.getLabel(), diagnosis.shortCNet, context);
			/**
			 * Add connection from the original net to this copy to the
			 * connection pool. As a result, this copy might be found next time.
			 */
			context.addConnection(new ShortCircuitedNetConnection(diagnosis.net, diagnosis.shortCNet, finalMarking));
			context.getProvidedObjectManager().createProvidedObject(
					"Initial marking of " + diagnosis.shortCNet.getLabel(), initialMarking, context);
			context.addConnection(new InitialMarkingConnection(diagnosis.shortCNet, initialMarking));
		}
	}

	/**
	 * Checks whether the net is a WF-net, using the assumptions.
	 * 
	 * @return WFNET if the net is a WF-net, NOWFNET if the net is not a WF-net,
	 *         DONE otherwise.
	 * @throws Exception
	 */
	private WoflanState diagnoseWFNet() {
		/**
		 * First check the assumptions.
		 */
		if (assumptions.contains(WoflanState.WFNET)) {
			return WoflanState.WFNET;
		} else if (assumptions.contains(WoflanState.NOWFNET)) {
			return WoflanState.NOWFNET;
		}
		/**
		 * No assumptions on this. Proceed.
		 */
		WoflanState state = WoflanState.DONE;
		if (WorkflowNetUtils.isValidWFNet(diagnosis.net)) {
			/**
			 * The net is a workflow net. Short-circuit it.
			 */
			shortCircuit();
			state = WoflanState.WFNET;
		} else {
			/**
			 * The net is not a workflow net. Diagnose the reason.
			 */
			SortedSet<Place> sourcePlaces = WorkflowNetUtils.getSourcePlaces(diagnosis.net);
			if (sourcePlaces.size() != 1) {
				diagnosis.setSourcePlaces(context, sourcePlaces);
			}
			SortedSet<Place> sinkPlaces = WorkflowNetUtils.getSinkPlaces(diagnosis.net);
			if (sinkPlaces.size() != 1) {
				diagnosis.setSinkPlaces(context, sinkPlaces);
			}
			if ((sourcePlaces.size() == 1) && (sinkPlaces.size() == 1)) {
				SortedSet<PetrinetNode> nodes = WorkflowNetUtils.getUnconnectedNodes(diagnosis.net);
				if (nodes.size() != 0) {
					diagnosis.setUnconnectedNodes(context, nodes);
				}
			}
			state = WoflanState.NOWFNET;
		}
		return state;
	}

	private WoflanState diagnoseSCover() throws ConnectionCannotBeObtained {
		/**
		 * Check assumptions.
		 */
		if (assumptions.contains(WoflanState.SCOVER)) {
			return WoflanState.SCOVER;
		} else if (assumptions.contains(WoflanState.NOSCOVER)) {
			return WoflanState.NOSCOVER;
		}
		WoflanState state = WoflanState.DONE;
		SComponentSet sComponents;
		/**
		 * Get the S-components.
		 */
		sComponents = context.tryToFindOrConstructFirstObject(SComponentSet.class, SComponentConnection.class,
				AbstractComponentSetConnection.COMPONENTSET, diagnosis.shortCNet);

		NotSCoveredNodesSet nodesSet = null;
		/**
		 * Try whether the set of not-S-covered nodes has already been computed.
		 */
		try {
			for (NotSCoveredNodesConnection conn : cm.getConnections(NotSCoveredNodesConnection.class, context,
					diagnosis.shortCNet)) {
				nodesSet = conn.getObjectWithRole(NotSCoveredNodesConnection.NODES);

			}
		} catch (Exception e) {
		}
		if (nodesSet == null) {
			/**
			 * No it is not. Compute it now.
			 */
			SortedSet<PetrinetNode> coveredNodes = new TreeSet<PetrinetNode>();
			for (SortedSet<PetrinetNode> nodes : sComponents) {
				coveredNodes.addAll(nodes);
			}
			SortedSet<PetrinetNode> uncoveredNodes = new TreeSet<PetrinetNode>(diagnosis.shortCNet.getPlaces());
			uncoveredNodes.addAll(diagnosis.shortCNet.getTransitions());
			uncoveredNodes.removeAll(coveredNodes);
			nodesSet = new NotSCoveredNodesSet();
			nodesSet.add(uncoveredNodes);
			context.addConnection(new NotSCoveredNodesConnection(diagnosis.shortCNet, nodesSet));
		}
		if (nodesSet.isEmpty() || nodesSet.iterator().next().isEmpty()) {
			state = WoflanState.SCOVER;
		} else {
			state = WoflanState.NOSCOVER;
			diagnosis.setNotSCoveredNodes(nodesSet);
		}
		return state;
	}

	private WoflanState diagnosePCover() throws ConnectionCannotBeObtained {
		/**
		 * Check assumptions.
		 */
		if (assumptions.contains(WoflanState.PCOVER)) {
			return WoflanState.PCOVER;
		} else if (assumptions.contains(WoflanState.NOPCOVER)) {
			return WoflanState.NOPCOVER;
		}
		WoflanState state = WoflanState.DONE;
		PlaceInvariantSet pInvariantSet;
		/**
		 * Get the place invariants.
		 */
		pInvariantSet = context.tryToFindOrConstructFirstObject(PlaceInvariantSet.class,
				PlaceInvariantConnection.class, AbstractInvariantMarkingConnection.INVARIANTMARKING,
				diagnosis.shortCNet);

		NotPCoveredNodesSet uncoveredNodes = null;
		try {
			for (NotPCoveredNodesConnection conn : cm.getConnections(NotPCoveredNodesConnection.class, context,
					diagnosis.shortCNet)) {
				uncoveredNodes = conn.getObjectWithRole(NotPCoveredNodesConnection.NODES);

			}
		} catch (Exception e) {
		}
		if (uncoveredNodes == null) {
			SortedSet<PetrinetNode> coveredNodes = new TreeSet<PetrinetNode>();
			for (MultiSet<Place> pInvariant : pInvariantSet) {
				coveredNodes.addAll(pInvariant.baseSet());
			}
			SortedSet<Place> uncoveredPlaces = new TreeSet<Place>(diagnosis.shortCNet.getPlaces());
			uncoveredPlaces.removeAll(coveredNodes);
			uncoveredNodes = new NotPCoveredNodesSet();
			uncoveredNodes.add(uncoveredPlaces);
			context.addConnection(new NotPCoveredNodesConnection(diagnosis.shortCNet, uncoveredNodes));
		}
		if (uncoveredNodes.isEmpty() || (uncoveredNodes.iterator().next().size() == 0)) {
			state = WoflanState.PCOVER;
		} else {
			diagnosis.setNotPCoveredNodes(uncoveredNodes);
			state = WoflanState.NOPCOVER;
		}
		return state;
	}

	private WoflanState diagnoseDeadness() throws ConnectionCannotBeObtained {
		/**
		 * Check assumptions.
		 */
		if (assumptions.contains(WoflanState.DEAD)) {
			return WoflanState.DEAD;
		} else if (assumptions.contains(WoflanState.NOTDEAD)) {
			return WoflanState.NOTDEAD;
		}
		WoflanState state = WoflanState.DONE;
		DeadTransitionsSet deadTransitionsSet;
		/**
		 * Get the dead transitions.
		 */
		deadTransitionsSet = context.tryToFindOrConstructFirstObject(DeadTransitionsSet.class,
				DeadTransitionsConnection.class, DeadTransitionsConnection.TRANSITIONS, diagnosis.shortCNet,
				initialMarking, diagnosis.semantics);

		if (deadTransitionsSet.isEmpty() || deadTransitionsSet.iterator().next().isEmpty()) {
			state = WoflanState.NOTDEAD;
		} else {
			state = WoflanState.DEAD;
			diagnosis.setDeadTransitions(deadTransitionsSet);
		}
		return state;
	}

	private WoflanState diagnoseLiveness() throws ConnectionCannotBeObtained {
		/**
		 * Check assumptions.
		 */
		if (assumptions.contains(WoflanState.LIVE)) {
			return WoflanState.LIVE;
		} else if (assumptions.contains(WoflanState.NOTLIVE)) {
			return WoflanState.NOTLIVE;
		}
		WoflanState state = WoflanState.DONE;
		NonLiveTransitionsSet nonLiveTransitionsSet;
		/**
		 * Try whether the set of non-live transitions has already been
		 * computed.
		 */
		nonLiveTransitionsSet = context.tryToFindOrConstructFirstObject(NonLiveTransitionsSet.class,
				NonLiveTransitionsConnection.class, NonLiveTransitionsConnection.TRANSITIONS, diagnosis.shortCNet,
				initialMarking, diagnosis.semantics);

		if (nonLiveTransitionsSet.isEmpty() || nonLiveTransitionsSet.iterator().next().isEmpty()) {
			state = WoflanState.LIVE;
		} else {
			state = WoflanState.NOTLIVE;
			NonLiveSequences nonLiveSequences;
			/**
			 * Get the non-live sequences.
			 */
			nonLiveSequences = context.tryToFindOrConstructFirstObject(NonLiveSequences.class,
					NonLiveSequencesConnection.class, NonLiveSequencesConnection.SEQUENCES, diagnosis.shortCNet,
					initialMarking, diagnosis.semantics, finalMarking);

			diagnosis.setNotLiveTransitions(nonLiveTransitionsSet, nonLiveSequences);
		}
		return state;
	}

	private WoflanState diagnoseBoundedness() throws ConnectionCannotBeObtained {
		/**
		 * Check assumptions.
		 */
		if (assumptions.contains(WoflanState.BOUNDED)) {
			return WoflanState.BOUNDED;
		} else if (assumptions.contains(WoflanState.NOTBOUNDED)) {
			return WoflanState.NOTBOUNDED;
		}
		WoflanState state = WoflanState.DONE;
		UnboundedPlacesSet unboundedPlacesSet = null;

		unboundedPlacesSet = context.tryToFindOrConstructFirstObject(UnboundedPlacesSet.class,
				UnboundedPlacesConnection.class, UnboundedPlacesConnection.PLACES, diagnosis.shortCNet, initialMarking,
				diagnosis.semantics);

		if (unboundedPlacesSet.isEmpty() || unboundedPlacesSet.iterator().next().isEmpty()) {
			state = WoflanState.BOUNDED;
		} else {
			state = WoflanState.NOTBOUNDED;
			UnboundedSequences unboundedSequences;
			/**
			 * Get the unbounded sequences.
			 */
			unboundedSequences = context.tryToFindOrConstructFirstObject(UnboundedSequences.class,
					UnboundedSequencesConnection.class, UnboundedSequencesConnection.SEQUENCES, diagnosis.shortCNet,
					initialMarking, diagnosis.semantics);

			diagnosis.setUnboundedPlaces(unboundedPlacesSet, unboundedSequences);
		}
		return state;
	}

	private WoflanState diagnoseFC() throws ConnectionCannotBeObtained {
		/**
		 * Check assumptions.
		 */
		if (assumptions.contains(WoflanState.FREECHOICE)) {
			return WoflanState.FREECHOICE;
		} else if (assumptions.contains(WoflanState.NONFREECHOICE)) {
			return WoflanState.NONFREECHOICE;
		}
		WoflanState state = WoflanState.DONE;
		NonExtendedFreeChoiceClustersSet clusterSet;
		/**
		 * Get the non-extended-free-choice clusters.
		 */
		clusterSet = context.tryToFindOrConstructFirstObject(NonExtendedFreeChoiceClustersSet.class,
				NonExtendedFreeChoiceClustersConnection.class, NonExtendedFreeChoiceClustersConnection.CLUSTERS,
				diagnosis.shortCNet);

		if (clusterSet.isEmpty()) {
			state = WoflanState.FREECHOICE;
		} else {
			state = WoflanState.NONFREECHOICE;
			diagnosis.setNotFreeChoice(clusterSet);
		}
		return state;
	}

	private WoflanState diagnosePTHandles() throws ConnectionCannotBeObtained {
		/**
		 * Check assumptions.
		 */
		if (assumptions.contains(WoflanState.WELLPTHANDLED)) {
			return WoflanState.WELLPTHANDLED;
		} else if (assumptions.contains(WoflanState.NOTWELLPTHANDLED)) {
			return WoflanState.NOTWELLPTHANDLED;
		}
		WoflanState state = WoflanState.DONE;
		PTHandles handles;
		/**
		 * Get the PT-handles.
		 */
		handles = context.tryToFindOrConstructFirstObject(PTHandles.class, PTHandleConnection.class,
				AbstractNetHandleConnection.HANDLES, diagnosis.shortCNet);

		if (handles.isEmpty()) {
			state = WoflanState.WELLPTHANDLED;
		} else {
			state = WoflanState.NOTWELLPTHANDLED;
			diagnosis.setPTHandles(handles);
		}
		return state;
	}

	private WoflanState diagnoseTPHandles() throws ConnectionCannotBeObtained {
		/**
		 * Check assumptions.
		 */
		if (assumptions.contains(WoflanState.WELLTPHANDLED)) {
			return WoflanState.WELLTPHANDLED;
		} else if (assumptions.contains(WoflanState.NOTWELLTPHANDLED)) {
			return WoflanState.NOTWELLTPHANDLED;
		}
		WoflanState state = WoflanState.DONE;
		TPHandles handles;
		/**
		 * Get the TP-handles.
		 */
		handles = context.tryToFindOrConstructFirstObject(TPHandles.class, TPHandleConnection.class,
				AbstractNetHandleConnection.HANDLES, diagnosis.shortCNet);

		if (handles.isEmpty()) {
			state = WoflanState.WELLTPHANDLED;
		} else {
			state = WoflanState.NOTWELLTPHANDLED;
			diagnosis.setTPHandles(handles);
		}
		return state;
	}
}
